pro part_a, search_tag, out_file, pad=pad

; Part_A computes the sums of n*n n*p n*r p*p p*r r*r for one ramp.
; These sums are combined by part_B to produce reference gain coefficients
; which can be applied using part_C.

; INPUT:
; search_tag = string with wildcard to identify the frames (FITS files) to be processed.
; OPTIONAL INPUT:
; out_file   = output filename [default = 'part_A_sums.xdr'] 

; OUTPUT is saved in "part_A_sums.xdr" in the current working directory.
;-------------------------------------------------------------------------------
st = systime(1)
file = file_search(search_tag,count=nfiles)
if (n_elements(out_file) ne 1) then out_file = file_basename(file,'.fits')+'_part_A_sums.xdr'

; Read header to get data setup
head = headfits(file[0])
scipix_n = sxpar(head,'NRS_NORM')
refpix_r = sxpar(head,'NRS_REF')
print,'SCIPIX_N = ',SCIPIX_N,'   REFPIX_R = ',REFpIX_R

if (n_elements(gain)) ne 1 then begin 
  case (sxpar(head,'SCA_ID')) of 
     491  : gain = 1.0  ; new detectors May 2014 ; changes 2PM may 14
     492  : gain = 1.11
;     491  : gain = 1.24 ; old detectors
;     492  : gain = 1.21
     else : gain = 1.34
  endcase
endif
;gain = 1.24 ; electrons/DN Array 491
print,'GAIN = ',gain

; This is the effective number of pixels sampled during the pause at the end of each row. 
; The padding is needed to preserve phase of temporally periodic signals. 
; Still good for prototype_v2, pad = 12
; Jan 2013, prototype_v3 pad=8
if (n_elements(pad) eq 0) then pad = 8 
row = (scipix_n+refpix_r+2)*(512/scipix_n)+pad

print,'mark1: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

; read data
data0 = float(readfits(file,/silent))
;data0 += (data0 lt 0)*2L^16
data0 *= gain

s = size(data0)
nn = lindgen(s[3])
nn -= mean(nn) ; subtract mean(nn) to minimize correlation of slope and offset

hnorm = lindgen(512)+refpix_r*((lindgen(512)+scipix_n/2) / scipix_n) ; column indices of normal pixels ; ***
href = lindgen(512/scipix_n*refpix_r)+scipix_n*(lindgen(512/scipix_n*refpix_r) / refpix_r)+scipix_n/2 ; column indices of reference pixels ; ***
hnorm1 = lindgen(512)+(refpix_r+2)*((lindgen(512)+scipix_n/2) / scipix_n) ; column indices of normal pixels ; ***
href1 = lindgen(512/scipix_n*refpix_r)+(scipix_n+2)*(lindgen(512/scipix_n*refpix_r) / refpix_r)+(scipix_n/2+1) ; column indices of reference pixels ; ***

; remove linear trends per pixel (gains, offsets)
data0 = reform(data0,s[1]*s[2],s[3],/over)
sxy = [nn##data0[0:1999,*],nn##data0[2000:*,*]] ; nn##data ; For obscure IDL reasons, this caluclation is faster when done in two parts.
sx = total(nn)
sxx = total(nn^2)
sy = [(nn*0+1)##data0[0:1999,*],(nn*0+1)##data0[2000:*,*]] ; total(data,2)
;m = (s[3]*sxy-sx*sy)/(s[3]*sxx-sx^2) ; ***
b = (sy*sxx-sxy*sx)/(s[3]*sxx-sx^2) ; ***
;  for i=0,s[3]-1 do data0[*,i] -= nn[i]*m+b  ; subtract slope and offset
for i=0,s[3]-1 do data0[*,i] -= b ; subtract the offset only
sxy = 0
sy = 0
b = 0
sig_data = sqrt([(nn*0+1)##data0[0:1999,*]^2,(nn*0+1)##data0[2000:*,*]^2]/(s[3]-1)) ; mean = 0 after ramp subtraction
sig_data = reform(sig_data,s[1],s[2],/over)
data0 = reform(data0,s[1],s[2],s[3],/over)
print,'mark2: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

  ; mask outliers with especially high or low stddev
!p.multi = [0,2,4]
nx5 = s[1]/5 ; all 5 outputs now
rev = reverse(lindgen(nx5))
mask = fltarr(s[1],s[2])
; set reference output to no masking
mask[0:nx5-1,*] = 1b
nmask = mask
for i=1,4 do nmask[hnorm+i*nx5,4:2043] = 1
nmask[nx5:nx5+3,*] = 0
nmask[s[1]-4:*,*] = 0
foreach i,[1,2,3,4] do begin
  msk = fltarr(nx5,s[2])
  hhnorm = where(nmask[i*nx5:(i+1)*nx5-1,*],compl=hhref)
  histogauss,(sig_data[i*nx5:(i+1)*nx5-1,*]-sig_data[0:nx5-1,*])[hhnorm],bb   ; HISTOGAUSS is in the IDLASTRO library at http://idlastro.gsfc.nasa.gov
  msk[hhnorm] = abs((sig_data[i*nx5:(i+1)*nx5-1,*]-sig_data[0:nx5-1,*])[hhnorm]-bb[1]) lt (4*bb[2])
  histogauss,(sig_data[i*nx5:(i+1)*nx5-1,*]-sig_data[0:nx5-1,*])[hhref],bb   ; HISTOGAUSS is in the IDLASTRO library at http://idlastro.gsfc.nasa.gov
  msk[hhref] = abs((sig_data[i*nx5:(i+1)*nx5-1,*]-sig_data[0:nx5-1,*])[hhref]-bb[1]) lt (4*bb[2])
  mask[i*nx5:(i+1)*nx5-1,*] = msk
endforeach
!p.multi = 0

  ; expand mask to cover x,y neighbor pixels.
  mask *= shift(mask,-1,0)*shift(mask,1,0)*shift(mask,0,-1)*shift(mask,0,1)
  ; reset reference output to no masking
  mask[0:nx5-1,*] = 1b

  ; apply mask to data and reform for later
  for i=0,s[3]-1 do data0[*,*,i] *= mask
  mask0 = mask
  mask0 = reform(mask0,nx5,5,s[2],/over)
  mask0 = transpose(mask0,[0,2,1])
  mask0[*,*,0] = mask0[rev,*,0]
  mask0[*,*,2] = mask0[rev,*,2]
  mask0[*,*,4] = mask0[rev,*,4]
  
  print,'mark3: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

  ; reform the data
  data0 = reform(data0,nx5,5,s[2],s[3],/over)
  data0 = transpose(data0,[0,2,3,1])
  data0[*,*,*,0] = data0[rev,*,*,0]
  data0[*,*,*,2] = data0[rev,*,*,2]
  data0[*,*,*,4] = data0[rev,*,*,4]

  ; convert to time sequences of normal pixels and reference pixels.
  d0 = fltarr(nx5+pad+2*(512/scipix_n),s[2],s[3],5)
  d0[hnorm1,*,*,*] = data0[hnorm,*,*,*]
  d0[href1,*,*,*] = data0[href,*,*,*]
  data0 = temporary(d0)

;<<<<< Fitting and removal of slopes per frame to remove issues at frame boundaries.
  time = findgen(row,s[2])
  row4plus4 = [0,1,2,3,2044,2045,2046,2047]
  time = time-mean(time)
  ab_1 = fltarr(2,s[3],5)  ; for normal pixels and reference output
  ab_2 = fltarr(2,s[3],5)  ; for windowed reference pixels
  ab_3 = fltarr(2,s[3],5)  ; for top+bottom ref pixel rows
  for i=0,4 do begin 
    for k=0,s[3]-1 do begin
;      ab_1[*,k,i] = linfit((time[hnorm1,4:s[2]-5])[where(data0[hnorm1,4:s[2]-5,k,i] ne 0)],(data0[hnorm1,4:s[2]-5,k,i])[where(data0[hnorm1,4:s[2]-5,k,i] ne 0)])
;      ab_2[*,k,i] = linfit((time[href1,4:s[2]-5])[where(data0[href1,4:s[2]-5,k,i] ne 0)],(data0[href1,4:s[2]-5,k,i])[where(data0[href1,4:s[2]-5,k,i] ne 0)])
      ab_3[*,k,i] = linfit((time[*,row4plus4])[where(data0[*,row4plus4,k,i] ne 0)],(data0[*,row4plus4,k,i])[where(data0[*,row4plus4,k,i] ne 0)])
    endfor
  endfor

for i=0,4 do begin 
  for k=0,s[3]-1 do data0[*,*,k,i] -= (time*ab_3[1,k,i]+ab_3[0,k,i])*(data0[*,*,k,i] ne 0)
endfor
time = 0
;<<<<<<<

  print,'mark4: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

; Use cosine weighted interpolation to replace 0.0 values at bad pixels and gaps. (initial guess)
  w = (sin((findgen(32)/32)*!pi))[1:*]
  for kk=0,4 do begin
    for jj=0L,s[3]-1 do begin
      dat = reform(data0[*,*,jj,kk],row*s[2])
      mask = float(dat ne 0)
      dat = reform(convol(dat,w,/edge_wrap)/convol(mask,w,/edge_wrap),row,s[2])
      data0[*,*,jj,kk] += dat*(1-mask)
    endfor
  endfor
  here = where(finite(data0,/nan),nhere)
  if (nhere ge 1) then data0[here] = 0
  print,'mark5: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Use Fourier filter/interpolation to replace  
; (a) bad pixel, gaps, and reference data in the time-ordered normal data
; (b) gaps and normal data in the time-ordered reference data
; This "improves" upon the cosine interpolation performed above.

; This process might also be made more memory efficient by using shared memory
; (SHMMAP, SHMVAR) instead of passing data0, but at this point in the code 
; avoiding a temprary duplication of the data is not critical. Shared memory
; might also improve speed may avoiding the copying, but this is probably only
; a matter of a few seconds.

; parameters for the filter to be used
  elen=110000L/(scipix_n+refpix_r+2)    ; length of apodization cosine filter
  blen=(512L+512L/scipix_n*(refpix_r+2)+pad)/(scipix_n+refpix_r+2)*s[2]/2-elen/2 ; max unfiltered frequency 

; construct the filter [1,cos,0,cos,1]
  a=(cos(findgen(elen)*!pi/elen)+1)/2 
  a=[replicate(1.,blen),a,fltarr(row*s[2]/2-2*blen-2*elen),reverse(a),replicate(1.,blen)]
  a=[a,reverse(shift(a,-1))]
  aa = a#replicate(1,s[3]) ; for application to the data
  afilter = a

; indices for keeping/shuffling reference pixels
  ht = indgen(scipix_n+refpix_r+2,512/scipix_n)
  ht[0:(scipix_n-refpix_r)/2,*]=-1
  ht[scipix_n/2+1+3*refpix_r/2:*,*]=-1 
  hs=ht 
  ht=reform(ht[where(ht ge 0)],2*refpix_r*512/scipix_n) ; t = like href1, but extended over gaps and 1st and last norm pix.
  hs[scipix_n/2+1-refpix_r/2:scipix_n/2+refpix_r+refpix_r/2,*]=hs[reform([transpose(reform(indgen(refpix_r),refpix_r/2,2)),transpose(reform(indgen(refpix_r),refpix_r/2,2))],refpix_r*2)+scipix_n/2+1,*]  ; WIRED for R=2^(int)
  hs=reform(hs[where(hs ge 0)],2*refpix_r*512/scipix_n) ; s = like href1, shuffled and replicated
  if (refpix_r mod 4 eq 2) then hs = reform(reverse(reform(hs,2,n_elements(hs)/2)),n_elements(hs))

  n_iter_ref = 0
  n_iter_norm = 3

; construct the reference data
  r0 = data0
  r0[*] = 0
  r0[ht,*,*,*] = data0[hs,*,*,*]

; construct the normal data
  oBridge = objarr(5)
  status = lonarr(5)
  for k=0,4 do begin
    oBridge[k] = OBJ_NEW("IDL_IDLBridge")
    if (!version.release ne '8.1') then oBridge[k]->Execute,"!path += ':'+expand_path('+~/arendt/idl')"
    oBridge[k]->Execute, '@' + PREF_GET('IDL_STARTUP')
    oBridge[k]->SetVar, "dd0", data0[*,*,*,k]
    oBridge[k]->SetVar, "k", k
    oBridge[k]->SetVar, "mask0", mask0
    oBridge[k]->SetVar, "row", row
    oBridge[k]->SetVar, "hnorm", hnorm
    oBridge[k]->SetVar, "hnorm1", hnorm1
    oBridge[k]->SetVar, "s", s
    oBridge[k]->SetVar, "aa", aa
    oBridge[k]->SetVar, "n_iter", n_iter_norm
  endfor
  for k=0,4 do oBridge[k]->Execute, "fft_interp_norm,dd0,k,mask0,row,hnorm,hnorm1,s,aa,n_iter", /nowait
  repeat begin
    for i=0,4 do status[i] = oBridge[i]->Status()
    print,systime(1)-st,status,string(13b),form='(f,5i8,a,$)'
    spawn,'sleep 1'
  endrep until (total(status) eq 0)
  print
  for k=0,4 do oBridge[k]->Execute, "sd = size(dd0)"
  for k=0,4 do oBridge[k]->Execute, "dd0 = reform(dd0,sd[1]*sd[2],sd[3],/over)"
  for k=0,4 do oBridge[k]->Execute, "dd0 = fft(dd0,dim=1,/over)",/nowait
  repeat begin
    for i=0,4 do status[i] = oBridge[i]->Status()
    print,systime(1)-st,status,string(13b),form='(f,5i8,a,$)'
    spawn,'sleep 1'
  endrep until (total(status) eq 0)
  print
  sd = size(data0)
  data0 = 0
  data0 = complexarr(sd[1]*sd[2],sd[3],sd[4])
  for k=0,4 do data0[*,*,k] = oBridge[k]->GetVar('dd0')
  for k=0,4 do obj_destroy,oBridge[k]
  print,'mark6: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

  r0 = reform(r0,sd[1]*sd[2],sd[3],5,/over)
  r0 = fft(r0,dim=1,/over)
  print,'mark7: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

  sum_nn_0 = total(abs(data0[*,*,1])^2,2)/(s[3]-1) ;***
  sum_nn_1 = total(abs(data0[*,*,2])^2,2)/(s[3]-1) ;***
  sum_nn_2 = total(abs(data0[*,*,3])^2,2)/(s[3]-1) ;***
  sum_nn_3 = total(abs(data0[*,*,4])^2,2)/(s[3]-1) ;***

  sum_rr_0 = total(abs(r0[*,*,1])^2,2)/(s[3]-1) ;***
  sum_rr_1 = total(abs(r0[*,*,2])^2,2)/(s[3]-1) ;***
  sum_rr_2 = total(abs(r0[*,*,3])^2,2)/(s[3]-1) ;***
  sum_rr_3 = total(abs(r0[*,*,4])^2,2)/(s[3]-1) ;***

  sum_oo   = total(abs(data0[*,*,0])^2,2)/(s[3]-1) ;***

  sum_nr_0 = total(data0[*,*,1]*conj(r0[*,*,1]),2)/(s[3]-1) ;***
  sum_nr_1 = total(data0[*,*,2]*conj(r0[*,*,2]),2)/(s[3]-1) ;***
  sum_nr_2 = total(data0[*,*,3]*conj(r0[*,*,3]),2)/(s[3]-1) ;***
  sum_nr_3 = total(data0[*,*,4]*conj(r0[*,*,4]),2)/(s[3]-1) ;***

  sum_no_0 = total(data0[*,*,1]*conj(data0[*,*,0]),2)/(s[3]-1) ;***
  sum_no_1 = total(data0[*,*,2]*conj(data0[*,*,0]),2)/(s[3]-1) ;***
  sum_no_2 = total(data0[*,*,3]*conj(data0[*,*,0]),2)/(s[3]-1) ;***
  sum_no_3 = total(data0[*,*,4]*conj(data0[*,*,0]),2)/(s[3]-1) ;***

  sum_ro_0 = total(r0[*,*,1]*conj(data0[*,*,0]),2)/(s[3]-1) ;***
  sum_ro_1 = total(r0[*,*,2]*conj(data0[*,*,0]),2)/(s[3]-1) ;***
  sum_ro_2 = total(r0[*,*,3]*conj(data0[*,*,0]),2)/(s[3]-1) ;***
  sum_ro_3 = total(r0[*,*,4]*conj(data0[*,*,0]),2)/(s[3]-1) ;***
  print,'mark8: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

save,file=out_file,sum_nn_0,sum_nn_1,sum_nn_2,sum_nn_3,sum_rr_0,sum_rr_1,sum_rr_2,sum_rr_3,$
sum_oo,sum_nr_0,sum_nr_1,sum_nr_2,sum_nr_3,sum_no_0,sum_no_1,sum_no_2,sum_no_3,sum_ro_0,sum_ro_1,sum_ro_2,sum_ro_3

end
